"""Various analysis protocols and standards for recorded underwater noise from ships."""
import functools
import numpy as np
from . import positional, propagation, _tools
import scipy.signal
import xarray as xr


class _DataWrapper:
    @staticmethod
    def _numop(binary=True):
        """Decorator for creating numerical wrappers.

        This decorator should be applied to functions that take self and other
        as the arguments, perform some simple numerical operations on the data,
        and returns the data. The decorated functions will check that the
        types make somewhat sense, extract the actual data,
        and wrap the output in the same class that the call was made on.
        """
        if binary:
            def wrapper(func):
                @functools.wraps(func)
                def wraps(self, other):
                    if type(self) == type(other):
                        other = other.data
                    elif isinstance(other, _DataWrapper):
                        return NotImplemented

                    data = func(self.data, other)
                    obj = type(self)(data)
                    self._transfer_attributes(obj)
                    return obj
                return wraps
        else:
            def wrapper(func):
                @functools.wraps(func)
                def wraps(self):
                    data = func(self.data)
                    obj = type(self)(data)
                    self._transfer_attributes(obj)
                    return obj
                return wraps
        return wrapper

    def __init__(self, data):
        self.data = data

    def _transfer_attributes(self, other):
        """Copy attributes form self to other

        This is useful to when creating a new copy of the same instance
        but with new data. The intent is for subclasses to extend
        this function to preserve attributes of the class that
        are not stored within the data variable.
        Note that this does not create a new instance of the class,
        so `other` should already be instantiated with data.
        The typical scheme to create a new instance from a new datastructure
        is
        ```
        new = type(self)(data)
        self._transfer_attributes(new)
        return new
        ```
        """
        pass

    @_numop()
    def __add__(self, other):
        return self + other

    @_numop()
    def __radd__(self, other):
        return other + self

    @_numop()
    def __sub__(self, other):
        return self - other

    @_numop()
    def __rsub__(self, other):
        return other - self

    @_numop()
    def __mul__(self, other):
        return self * other

    @_numop()
    def __rmul__(self, other):
        return other * self

    @_numop()
    def __truediv__(self, other):
        return self / other

    @_numop()
    def __rtruediv__(self, other):
        return other / self

    @_numop()
    def __floordiv__(self, other):
        return self // other

    @_numop()
    def __rfloordiv__(self, other):
        return other // self

    @_numop()
    def __pow__(self, other):
        return self ** other

    @_numop()
    def __rpow__(self, other):
        return other ** self

    @_numop()
    def __mod__(self, other):
        return self % other

    @_numop()
    def __rmod__(self, other):
        return other % self

    @_numop(binary=False)
    def __neg__(self):
        return -self

    @_numop(binary=False)
    def __abs__(self):
        return abs(self)


class TimeData(_DataWrapper):
    @staticmethod
    def _with_time_vector(data, start_time, samplerate):
        if samplerate is None:
            return data
        if start_time is None:
            if 'time' in data.coords:
                start_time = data.time[0].item()
            start_time = 'now'
        n_samples = data.sizes['time']
        start_time = positional.time_to_np(start_time)
        offsets = np.arange(n_samples) * 1e9 / samplerate
        time = start_time + offsets.astype('timedelta64[ns]')
        return data.assign_coords(time=('time', time, {'rate': samplerate}))

    def __init__(self, data, start_time=None, samplerate=None, dims=None, coords=None, **kwargs):
        if not isinstance(data, xr.DataArray):
            if dims is None:
                if data.ndim == 1:
                    dims = 'time'
                else:
                    raise ValueError(f'Cannot guess dimensions for time data with {data.ndim} dimensions')
            data = xr.DataArray(data, dims=dims)

        if coords is not None:
            data = data.assign_coords(**{name: coord for (name, coord) in coords.items() if name not in {'time'}})
        data = self._with_time_vector(data, samplerate=samplerate, start_time=start_time)
        super().__init__(data, **kwargs)
        # self.data = data

    @property
    def time(self):
        return self.data.time

    @property
    def samplerate(self):
        return self.data.time.rate

    @property
    def time_window(self):
        # Calculating duration from number and rate means the stop points to the sample after the last,
        # which is more intuitive when considering signal durations etc.
        return positional.TimeWindow(
            start=self.data.time.data[0],
            duration=self.data.sizes['time'] / self.samplerate,
        )

    def subwindow(self, time=None, /, *, start=None, stop=None, center=None, duration=None, extend=None):
        original_window = self.time_window
        new_window = original_window.subwindow(time, start=start, stop=stop, center=center, duration=duration, extend=extend)
        if isinstance(new_window, positional.TimeWindow):
            start = (new_window.start - original_window.start).total_seconds()
            stop = (new_window.stop - original_window.start).total_seconds()
            # Indices assumed to be seconds from start
            start = np.math.floor(start * self.samplerate)
            stop = np.math.ceil(stop * self.samplerate)
            idx = slice(start, stop)
        else:
            idx = (new_window - original_window.start).total_seconds()
            idx = round(idx * self.samplerate)

        selected_data = self.data.isel(time=idx)
        new = type(self)(selected_data)
        self._transfer_attributes(new)
        return new


class FrequencyData(_DataWrapper):
    @staticmethod
    def _with_frequency_bandwidth_vectors(data, frequency, bandwidth):
        if frequency is None:
            return data
        coords = {'frequency': frequency}
        if bandwidth is not None:
            coords['bandwidth'] = ('frequency', np.broadcast_to(bandwidth, np.shape(frequency)))
        return data.assign_coords(coords)

    def __init__(self, data, frequency=None, bandwidth=None, scaling="power spectral density", dims=None, coords=None, **kwargs):
        if not isinstance(data, xr.DataArray):
            if dims is None:
                if data.ndim == 1:
                    dims = 'frequency'
                else:
                    raise ValueError(f'Cannot guess dimensions for frequency data with {data.ndim} dimensions')
            data = xr.DataArray(data, dims=dims)
        if coords is not None:
            data = data.assign_coords(**{name: coord for (name, coord) in coords.items() if name not in {"frequency", "bandwidth"}})
        data = self._with_frequency_bandwidth_vectors(data, frequency=frequency, bandwidth=bandwidth)
        super().__init__(data, **kwargs)
        # self.data = data
        self.scaling = scaling

    @property
    def frequency(self):
        return self.data.frequency

    def _transfer_attributes(self, other):
        super()._transfer_attributes(other)
        try:
            other.scaling = self.scaling
        except AttributeError:
            pass

    def as_power_spectrum(self):
        if "density" in self.scaling.lower():
            new = type(self)(self.data * self.data.bandwith)
            self._transfer_attributes(new)
            new.scaling = "power spectrum"
        else:
            new = type(self)(self.data.copy())
            self._transfer_attributes(new)
        return new

    def as_power_spectral_density(self):
        if "density" not in self.scaling.lower():
            new = type(self)(self.data / self.data.bandwith)
            self._transfer_attributes(new)
            new.scaling = "power spectral density"
        else:
            new = type(self)(self.data.copy())
            self._transfer_attributes(new)
        return new

    def estimate_bandwidth(self):
        frequency = np.asarray(self.frequency)
        # Check if the frequency array seems linearly or logarithmically spaced
        if frequency[0] == 0:
            diff = frequency[2:] - frequency[1:-1]
            frac = frequency[2:] / frequency[1:-1]
        else:
            diff = frequency[1:] - frequency[:-1]
            frac = frequency[1:] / frequency[:-1]
        diff_err = np.std(diff) / np.mean(diff)
        frac_err = np.std(frac) / np.mean(frac)
        # Note: if there are three values and the first is zero, the std is 0 for both.
        # The equals option makes us end up in the linear frequency case.
        if diff_err <= frac_err:
            # Central differences, with forwards and backwards at the ends
            central = (frequency[2:] - frequency[:-2]) / 2
            first = frequency[1] - frequency[0]
            last = frequency[-1] - frequency[-2]
        else:
            # upper edge is at sqrt(f_{l+1} * f_l), lower edge is at sqrt(f_{l-1} * f_l)
            # the difference simplifies as below.
            central = (frequency[2:]**0.5 - frequency[:-2]**0.5) * frequency[1:-1]**0.5
            # extrapolating to one bin below lowest and one above highest using constant ratio
            # the expression above then simplifies to the expressions below
            first = (frequency[1] - frequency[0]) * (frequency[0] / frequency[1])**0.5
            last = (frequency[-1] - frequency[-2]) * (frequency[-1] / frequency[-2])**0.5
        bandwidth = np.concatenate([[first], central, [last]])
        return xr.DataArray(bandwidth, coords={'frequency': self.frequency})


class TimeFrequencyData(TimeData, FrequencyData):
    def __init__(self, data, start_time=None, samplerate=None, frequency=None, bandwidth=None, dims=None, coords=None, **kwargs):
        if not isinstance(data, xr.DataArray):
            if dims is None:
                raise ValueError('Cannot guess dimensions for time-frequency data')
            data = xr.DataArray(data, dims=dims)
        if coords is not None:
            data = data.assign_coords(**{name: coord for (name, coord) in coords.items() if name not in {"time", "frequency", "bandwidth"}})
        # data = self._with_time_vector(data, samplerate=samplerate, start_time=start_time)
        # data = self._with_frequency_bandwidth_vectors(data, frequency=frequency, bandwidth=bandwidth)
        super().__init__(
            data,
            start_time=start_time, samplerate=samplerate,
            frequency=frequency, bandwidth=bandwidth,
            **kwargs
        )
        # self.data = data


class Transit:
    def __init__(self, recording, track):
        start = max(recording.time_window.start, track.time_window.start)
        stop = min(recording.time_window.stop, track.time_window.stop)

        self.recording = recording.subwindow(start=start, stop=stop)
        self.track = track.subwindow(start=start, stop=stop)

    @property
    def time_window(self):
        rec_window = self.recording.time_window
        track_window = self.track.time_window
        return positional.TimeWindow(start=max(rec_window.start, track_window.start), stop=min(rec_window.stop, track_window.stop))

    def subwindow(self, time=None, /, *, start=None, stop=None, center=None, duration=None):
        subwindow = self.time_window.subwindow(time, start=start, stop=stop, center=center, duration=duration)
        rec = self.recording.subwindow(subwindow)
        track = self.track.subwindow(subwindow)
        return type(self)(recording=rec, track=track)


def dB(x, power=True, safe_zeros=True, ref=1):
    '''Calculate the decibel of an input value

    Parameters
    ----------
    x : numeric
        The value to take the decibel of
    power : boolean, default True
        Specifies if the input is a power-scale quantity or a root-power quantity.
        For power-scale quantities, the output is 10 log(x), for root-power quantities the output is 20 log(x).
        If there are negative values in a power-scale input, the handling can be controlled as follows:
        - `power='imag'`: return imaginary values
        - `power='nan'`: return nan where power < 0
        - `power=True`: as `nan`, but raises a warning.
    safe_zeros : boolean, default True
        If this option is on, all zero values in the input will be replaced with the smallest non-zero value.
    ref : numeric
        The reference unit for the decibel. Note that this should be in the same unit as the `x` input,
        e.g., if `x` is a power, the `ref` value might need squaring.
    '''
    if isinstance(x, _DataWrapper):
        new = dB(x.data, power=power, safe_zeros=safe_zeros, ref=ref)
        new = type(x)(new)
        x._transfer_attributes(new)
        return new

    if safe_zeros and np.size(x) > 1:
        nonzero = x != 0
        min_value = np.nanmin(abs(xr.where(nonzero, x, np.nan)))
        x = xr.where(nonzero, x, min_value)
    if power:
        if np.any(x < 0):
            if power == 'imag':
                return 10 * np.log10(x + 0j)
            if power == 'nan':
                return 10 * np.log10(xr.where(x > 0, x, np.nan))
        return 10 * np.log10(x / ref)
    else:
        return 20 * np.log10(np.abs(x) / ref)


class Spectrogram(TimeFrequencyData):
    """Calculates spectrograms

    The processing is done in stft frames determined by `frame_duration`, `frame_step`
    `frame_overlap`, and `frequency_resolution`. At least one of "duration", "step",
    or "resolution" has to be given, see `time_frame_settings` for further details.

    Parameters
    ----------
    data : TimeData, optional
        The time data from which to calculate the spectrogram.
        Omit this to create a callable object for later evaluation.
    frame_duration : float
        The duration of each stft frame, in seconds.
    frame_step : float
        The time step between stft frames, in seconds.
    frame_overlap : float, default 0.5
        The overlap factor between stft frames. A negative value leaves
        gaps between frames.
    frequency_resolution : float
        A frequency resolution to aim for. Only used if `frame_duration` is not given.
    fft_window : str, default="hann"
        The shape of the window used for the stft.
    """
    def __init__(
            self,
            data=None,
            frame_duration=None,
            frame_step=None,
            frame_overlap=0.5,
            frequency_resolution=None,
            fft_window="hann",
            **kwargs
        ):
        try:
            self.frame_settings = time_frame_settings(
                duration=frame_duration,
                step=frame_step,
                resolution=frequency_resolution,
                overlap=frame_overlap,
            ) | {"window": fft_window}
        except ValueError:
            pass

        if isinstance(data, TimeData):
            # __call__ will create an object that we only use the .data from
            data = self(data).data
        if data is not None:
            super().__init__(data, **kwargs)

    def __call__(self, time_data):
        if isinstance(time_data, type(self)):
            return time_data
        xr_data = time_data.data
        frame_samples = round(self.frame_settings["duration"] * time_data.samplerate)
        overlap_samples = round(self.frame_settings["duration"] * self.frame_settings["overlap"] * time_data.samplerate)
        f, t, Sxx = scipy.signal.spectrogram(
            x=xr_data.data,  # Gets the numpy array
            fs=time_data.samplerate,
            window=self.frame_settings["window"],
            nperseg=frame_samples,
            noverlap=overlap_samples,
            axis=xr_data.dims.index('time'),
        )
        dims = list(xr_data.dims)
        dims[dims.index('time')] = 'frequency'
        dims.append('time')
        new = type(self)(
            data=Sxx.copy(),  # Using a copy here is a performance improvement in later processing stages.
            # The array returned from the spectrogram function is the real part of the original stft, reshaped.
            # This means that the array takes twice the memory (the imaginary part is still around),
            # and it's not contiguous which slows down filtering a lot.
            samplerate=time_data.samplerate / (frame_samples - overlap_samples),
            start_time=time_data.time_window.start.add(seconds=t[0]),
            frequency=f,
            bandwidth=time_data.samplerate / frame_samples,
            dims=tuple(dims),
            coords=xr_data.coords,
        )
        self._transfer_attributes(new)
        return new

    def _transfer_attributes(self, obj):
        super()._transfer_attributes(obj)
        obj.frame_settings = self.frame_settings


class NthDecadeSpectrogram(TimeFrequencyData):
    """Calculates Nth-decade spectrograms.

    The processing is done in stft frames determined by `frame_duration`, `frame_step`
    `frame_overlap`, and `hybrid_resolution`. At least one of "duration", "step",
    or "resolution" has to be given, see `time_frame_settings` for further details.
    At least one of `lower_bound` and `hybrid_resolution` has to be given.
    Note that the `frame_duration` and `frame_step` can be auto-chosen from the overlap
    and required frequency resolution, either from `hybrid_resolution` or `lower_bound`.

    Parameters
    ----------
    data : TimeData or Spectrogram, optional
        The time data from which to calculate the spectrogram.
        Omit this to create a callable object for later evaluation.
    frame_duration : float
        The duration of each stft frame, in seconds.
    frame_step : float
        The time step between stft frames, in seconds.
    frame_overlap : float, default 0.5
        The overlap factor between stft frames. A negative value leaves
        gaps between frames.
    lower_bound : float
        The lowest frequency to include in the processing.
    upper_bound : float
        The highest frequency to include in the processing.
    hybrid_resolution : float
        A frequency resolution to aim for. Only used if `frame_duration` is not given.
    scaling : str, default="power spectral density"
        The scaling to use for the output.
        - "power spectral density" scales the output as a power spectral density.
        - "power spectrum" scales the output as the total power in each band.
    fft_window : str, default="hann"
        The shape of the window used for the stft.

    Raises
    ------
    ValueError
        If the processing settings are not compatible, e.g.,
        - frequency bands with bandwidth smaller than the frame duration allows,
        - no lower bound and no hybrid resolution.
    """
    def __init__(
        self,
        data=None,
        bands_per_decade=None,
        frame_step=None,
        frame_duration=None,
        frame_overlap=0.5,
        lower_bound=None,
        upper_bound=None,
        hybrid_resolution=None,
        # scaling="power spectral density",
        fft_window="hann",
        **kwargs
    ):
        try:
            if hybrid_resolution not in {True, False}:
                resolution = hybrid_resolution
            else:
                resolution = lower_bound * (10 ** (0.5 / bands_per_decade) - 10 ** (-0.5 / bands_per_decade))
            self.frame_settings = time_frame_settings(
                duration=frame_duration,
                step=frame_step,
                resolution=resolution,
                overlap=frame_overlap,
            ) | {"window": fft_window}
        except ValueError:
            pass

        self.bands_per_decade = bands_per_decade
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        # self.scaling = scaling
        self.hybrid_resolution = hybrid_resolution

        if isinstance(data, TimeData):
            # __call__ will create an object that we only use the .data from
            data = self(data).data
        if data is not None:
            super().__init__(data, **kwargs)

    def __call__(self, data):
        if isinstance(data, type(self)):
            return data
        if not isinstance(data, Spectrogram):
            data = Spectrogram(
                data,
                frame_step=self.frame_settings["step"],
                frame_duration=self.frame_settings["duration"],
                frame_overlap=self.frame_settings["overlap"],
                frequency_resolution=self.frame_settings["resolution"],
            )

        if self.hybrid_resolution:
            if self.hybrid_resolution is True:
                hybrid_resolution = 1 / data.frame_settings["duration"]
            else:
                hybrid_resolution = self.hybrid_resolution
            if hybrid_resolution * data.frame_settings["duration"] < 1:
                raise ValueError(
                    f'Hybrid filterbank with resolution of {hybrid_resolution:.2f} Hz '
                    f'cannot be calculated from temporal windows of {data.frame_settings["duration"]:.2f} s.'
                )
        else:
            hybrid_resolution = False
            lowest_bandwidth = self.lower_bound * (10 ** (0.5 / self.bands_per_decade) - 10 ** (-0.5 / self.bands_per_decade))
            if lowest_bandwidth * data.frame_settings["duration"] < 1:
                raise ValueError(
                    f'{self.bands_per_decade}th-decade filter band at {self.lower_bound:.2f} Hz with bandwidth of {lowest_bandwidth:.2f} Hz '
                    f'cannot be calculated from temporal windows of {data.frame_settings["duration"]:.2f} s.'
                )

        spec_xr = data.data.transpose("frequency", ...)
        # Get frequency centers for the new bands
        bands_per_decade = self.bands_per_decade
        log_band_scaling = 10 ** (0.5 / bands_per_decade)
        lower_bound = self.lower_bound or 0  # Prevent None or False.
        upper_bound = self.upper_bound or spec_xr.frequency.data[-1] / log_band_scaling
        if hybrid_resolution:
            # The frequency at which the logspaced bands cover at least one linspaced band
            minimum_bandwidth_frequency = hybrid_resolution / (log_band_scaling - 1 / log_band_scaling)
            first_log_idx = np.math.ceil(bands_per_decade * np.log10(minimum_bandwidth_frequency / 1e3))
            last_linear_idx = np.math.floor(minimum_bandwidth_frequency / hybrid_resolution)

            # Since the logspaced bands have pre-determined centers, we can't just start them after the linspaced bands.
            # Often, the bands will overlap at the minimum bandwidth frequency, so we look for the first band
            # that does not overlap, i.e., the upper edge of the last linspaced band is below the lower edge of the first
            # logspaced band
            while (last_linear_idx + 0.5) * hybrid_resolution > 1e3 * 10 ** ((first_log_idx - 0.5) / bands_per_decade):
                # Condition is "upper edge of last linear band is higher than lower edge of first logarithmic band"
                last_linear_idx += 1
                first_log_idx += 1

            if last_linear_idx * hybrid_resolution > upper_bound:
                last_linear_idx = np.math.floor(upper_bound / hybrid_resolution)
            first_linear_idx = np.math.ceil(lower_bound / hybrid_resolution)
        else:
            first_linear_idx = last_linear_idx = 0
            first_log_idx = np.round(bands_per_decade * np.log10(lower_bound / 1e3))

        last_log_idx = round(bands_per_decade * np.log10(upper_bound / 1e3))

        lin_centers = np.arange(first_linear_idx, last_linear_idx) * hybrid_resolution
        lin_lowers = lin_centers - 0.5 * hybrid_resolution
        lin_uppers = lin_centers + 0.5 * hybrid_resolution

        log_centers = 1e3 * 10 ** (np.arange(first_log_idx, last_log_idx + 1) / bands_per_decade)
        log_lowers = log_centers / log_band_scaling
        log_uppers = log_centers * log_band_scaling

        centers = np.concatenate([lin_centers, log_centers])
        lowers = np.concatenate([lin_lowers, log_lowers])
        uppers = np.concatenate([lin_uppers, log_uppers])

        # Aggregate input spectrum into bands
        spec_np = spec_xr.data
        banded_data = np.full(centers.shape + spec_np.shape[1:], np.nan)
        spectral_resolution = 1 / data.frame_settings["duration"]
        for idx, (l, u) in enumerate(zip(lowers, uppers)):
            l_idx = int(np.floor(l / spectral_resolution + 0.5))  # (l_idx - 0.5) * Δf = l
            u_idx = int(np.ceil(u / spectral_resolution - 0.5))  # (u_idx + 0.5) * Δf = u
            l_idx = max(l_idx, 0)
            u_idx = min(u_idx, spec_np.shape[0] - 1)

            if l_idx == u_idx:
                # This can only happen if both frequencies l and u are within the same fft bin.
                # Since we don't allow the fft bins to be larger than the output bins, we thus have the exact same band.
                banded_data[idx] = spec_np[l_idx]
            else:
                # weight edge bins by "(whole bin - what is not in the new band) / whole bin"
                # lower fft bin edge l_e = (l_idx - 0.5) * Δf
                # w_l = (Δf - (l - l_e)) / Δf = l_idx + 0.5 - l / Δf
                first_weight = l_idx + 0.5 - l / spectral_resolution
                # upper fft bin edge u_e = (u_idx + 0.5) * Δf
                # w_u = (Δf - (u_e - u)) / Δf = 0.5 - u_idx + u / Δf
                last_weight = u / spectral_resolution - u_idx + 0.5
                # Sum the components fully within the output bin `[l_idx + 1:u_idx]`, and weighted components partially in the band.
                this_band = (
                    spec_np[l_idx + 1 : u_idx].sum(axis=0)
                    + spec_np[l_idx] * first_weight
                    + spec_np[u_idx] * last_weight
                )
                banded_data[idx] = this_band * (spectral_resolution / (u - l))  # Rescale the power density.
        banded = type(self)(
            data=banded_data,
            start_time=spec_xr.time[0],
            samplerate=spec_xr.time.rate,
            frequency=centers,
            bandwidth=uppers - lowers,
            dims=spec_xr.dims,
            coords=spec_xr.coords,
        )
        if "density" not in self.scaling.lower():
            banded = banded.as_power_spectrum()
        self._transfer_attributes(banded)
        return banded

    # def total_power_in_bands(self):
    #     return self.data * self.data.bandwidth


class ShipLevel:
    @classmethod
    def analyze_transits(
        cls,
        *transits,
        filterbank=None,
        propagation_model=None,
        background_noise=None,
        transit_min_angle=None,
        transit_min_duration=None,
        transit_min_lengh=None,
    ):
        if filterbank is None:
            filterbank = NthDecadeSpectrogram(
                bands_per_decade=10,
                lower_bound=20,
                upper_bound=20_000,
                frame_step=1
            )

        if background_noise is None:
            def background_noise(received_power, **kwargs):
                return received_power

        if propagation_model is None:
            propagation_model = propagation.MlogR(m=20)

        if isinstance(propagation_model, propagation.PropagationModel):
            propagation_model = propagation_model.compensate_propagation

        results = []
        for transit in transits:
            if (transit_min_angle, transit_min_duration, transit_min_lengh) == (None, None, None):
                cpa_time = transit.track.closest_point(transit.recording.sensor)["time"].data
            else:
                segment = transit.track.aspect_segments(
                    reference=transit.recording.sensor,
                    angles=0,
                    segment_min_duration=transit_min_duration,
                    segment_min_angle=transit_min_angle,
                    segment_min_length=transit_min_lengh,
                )
                cpa_time = segment.time.sel(edge="center").data
                transit = transit.subwindow(segment)

            direction = transit.track.average_course('eight')
            time_data = transit.recording.time_data()
            received_power = filterbank(time_data)

            received_power = background_noise(received_power)  # TODO: Implement background correction wrappers
            # TODO: make the background correction wrappers store the snr alongside the received power?
            track = type(transit.track)(transit.track._data.interp(time=received_power.time))  # TODO: move this to a Track.resample method
            source_power = propagation_model(received_power=received_power, receiver=transit.recording.sensor, source=track)
            transit_time = (received_power.data["time"] - cpa_time) / np.timedelta64(1, "s")
            closest_to_cpa = np.abs(transit_time).argmin("time").item()
            segment = xr.DataArray(np.arange(transit_time.time.size) - closest_to_cpa, coords={"time": received_power.time})
            transit_results = xr.Dataset(
                data_vars=dict(
                    source_power=source_power.data,
                    latitude=track.latitude,
                    longitude=track.longitude,
                    transit_time=transit_time,
                ),
                coords=dict(
                    segment=segment,
                    direction=direction,
                )
            )
            transit_results["received_power"] = received_power.data
            if hasattr(received_power, "snr"):
                transit_results["snr"] = received_power.snr.data
            results.append(transit_results.swap_dims(time="segment").reset_coords("time"))
        results = xr.concat(results, "transit")
        return cls(results)

    def __init__(self, data):
        # data is xr.Dataset with source_level, received_level, snr
        # dims: segment, transit, sensor (optional?)
        # you could possible skip calculating some props eg. direction until you want them
        # coords: lat, lon, direction
        # more stuff: cpa time, cpa distance,
       self.data = data

    @property
    def source_power(self):
        return self.data["source_power"]

    @property
    def source_level(self):
        return dB(self.source_power, power=True)

    @property
    def received_power(self):
        return self.data["received_power"]

    @property
    def received_level(self):
        return dB(self.received_power, power=True)

    def mean(self, dims, **kwargs):
        return type(self)(self.data.mean(dims, **kwargs))


def bureau_veritas_source_spectrum(
    passages,
    propagation_model=None,
    background_noise=None,
    filterbank=None,
    aspect_angles=tuple(range(-45, 46, 5)),
    aspect_segment_length=100,
    aspect_segment_angle=None,
    aspect_segment_duration=None,
    passage_time_padding=10,
):
    if filterbank is None:
        filterbank = decidecade_filter(lower_bound=10, upper_bound=50_000, window_duration=1, overlap=0.5)
    if propagation_model is None:
        propagation_model = propagation.MlogR(m=20)

    if isinstance(propagation_model, propagation.PropagationModel):
        propagation_model = propagation_model.compensate_propagation

    if background_noise is None:
        def background_noise(received_power, **kwargs):
            return received_power

    passage_powers = []
    for passage_idx, passage in enumerate(passages):
        cpa = positional.closest_point(passage.recording.sensor, passage.track)
        time_segments = positional.aspect_segments(
            reference=passage.recording.sensor,
            track=passage.track,
            angles=aspect_angles,
            segment_min_length=aspect_segment_length,
            segment_min_angle=aspect_segment_angle,
            segment_min_duration=aspect_segment_duration,
        )
        time_start = positional.time_to_datetime(time_segments.time.min()).subtract(seconds=passage_time_padding)
        time_stop = positional.time_to_datetime(time_segments.time.max()).add(seconds=passage_time_padding)
        time_data = passage.recording.sampling.subwindow(start=time_start, stop=time_stop).time_data()
        received_power = filterbank(time_data)

        segment_powers = []
        for segment_idx, segment in time_segments.groupby('segment'):
            received_segment = received_power.sampling.subwindow(segment).reduce(np.mean, dim='time')
            source = passage.track.sampling.subwindow(segment.time.sel(edge='center'))

            compensated_segment = background_noise(
                received_segment,
                receiver=passage.recording.sensor,
                time=source.time,
            )
            source_segment = propagation_model(
                compensated_segment,
                receiver=passage.recording.sensor,
                source=source,
            )
            source_segment = source_segment.assign_coords(
                segment=segment.segment,
                latitude=source.latitude,
                longitude=source.longitude,
                time=source.time,
            )
            segment_powers.append(source_segment)

        segment_powers = xr.concat(segment_powers, dim='segment')
        passage_powers.append(segment_powers.assign_coords(cpa=cpa.distance))
    source_powers = xr.concat(passage_powers, dim='passage')
    return source_powers


def time_frame_settings(
    duration=None,
    step=None,
    overlap=None,
    resolution=None,
    num_frames=None,
    signal_length=None,
):
    """Calculates time frame overlap settings from various input parameters.

    Parameters
    ----------
    duration : float
        How long each frame is, in seconds
    step : float
        The time between frame starts, in seconds
    overlap : float
        How much overlap there is between the frames, as a fraction of the duration.
        If this is negative the frames will have extra space between them.
    resolution : float
        Desired frequency resolution in Hz. Equals `1/duration`
    num_frames : int
        The total number of frames in the signal
    signal_length : float
        The total length of the signal, in seconds

    Returns
    -------
    dict with keys
        - "duration"
        - "step"
        - "overlap"
        - "resolution"
        and if `signal_length` was given
        - "num_frames"
        - "signal_length"

    Raises
    ------
    ValueError
        If the inputs are not sufficient to determine the frame,
        or if priorities cannot be disambiguated.

    Notes
    -----
    The parameters will be used in the following priority:
    1. signal_length, num_frames
    2. step, duration
    3. resolution (only if duration not given)
    4. overlap (default 0)

    Each frame idx=[0, ..., num_frames - 1] has
    - start = idx * step
    - stop = idx * step + duration

    The last frame thus ends at (num_frames - 1) step + duration.
    The overlap relations are:
    - duration = step / (1 - overlap)
    - step = duration (1 - overlap)
    - overlap = 1 - step / duration

    This gives us the following total list of priorities:
    1) `signal_length` and `num_frames` are given
        a) `step` or `duration` (not both!) given
        b) `resolution` is given
        c) `overlap` is given
    2) `step` and `duration` given
    3) `step` given
        a) `resolution` is given
        b) `overlap` is given
    4) `duration` given (`resolution` is ignored, `overlap` is used)
    5) `resolution` given
    For cases 2-5, `num_frames` is calculated if `signal_length` is given,
    and a new truncated `signal_length` is returned.
    """
    if None not in (num_frames, signal_length):
        if None not in (duration, step):
            raise ValueError("Overdetermined time frames")
        elif step is not None:
            # We have the step, calculate the duration
            duration = signal_length - (num_frames - 1) * step
            overlap = 1 - step / duration
        elif duration is not None:
            # We have the duration, calculate the step
            step = (signal_length - duration) / (num_frames - 1)
            overlap = 1 - step / duration
        elif resolution is not None:
            duration = 1 / resolution
            step = (signal_length - duration) / (num_frames - 1)
            overlap = 1 - step / duration
        else:
            overlap = overlap or 0
            duration = signal_length / (num_frames + overlap - num_frames * overlap)
            step = duration * (1 - overlap)
    elif None not in (step, duration):
        overlap = 1 - step / duration
    elif step is not None:
        if resolution is not None:
            duration = 1 / resolution
            overlap = 1 - step / duration
        else:
            overlap = overlap or 0
            duration = step / (1 - overlap)
    elif duration is not None:
        overlap = overlap or 0
        step = duration * (1 - overlap)
    elif resolution is not None:
        duration = 1 / resolution
        overlap = overlap or 0
        step = duration * (1 - overlap)
    else:
        raise ValueError("Must give at least one of (`step`, `duration`, `resolution`) or the pair of `signal_length` and `num_frames`.")

    settings = {
        "duration": duration,
        "step": step,
        "overlap": overlap,
        "resolution": 1 / duration,
    }
    if signal_length is not None:
        num_frames = num_frames or np.math.floor((signal_length - duration) / step + 1)
        settings["num_frames"] = num_frames
        settings["signal_length"] = (num_frames - 1) * step + duration
    return settings


@_tools.prebind
def fft(time_signal, nfft=None):
    nfft = nfft or time_signal.sizes['time']
    return xr.apply_ufunc(
        np.fft.rfft,
        time_signal,
        input_core_dims=[['time']],
        output_core_dims=[['frequency']],
        kwargs={'n': nfft},
    ).assign_coords(frequency=np.fft.rfftfreq(nfft, 1 / time_signal.time.rate), time=time_signal.time[0]).rename(time='start_time')


@_tools.prebind
def ifft(spectrum, nfft=None):
    if nfft is None:
        is_odd = np.any(spectrum.isel(frequency=-1).data.imag)
        nfft = (spectrum.sizes['frequency'] - 1) * 2 + (1 if is_odd else 0)

    time_data = xr.apply_ufunc(
        np.fft.irfft,
        spectrum,
        input_core_dims=[['frequency']],
        output_core_dims=[['time']],
        kwargs={'n': nfft},
    ).drop_vars('start_time')
    samplerate = spectrum.frequency[1].item() * nfft
    if hasattr(spectrum, 'start_time'):
        return recordings.time_data(time_data, start_time=spectrum.start_time, samplerate=samplerate)
    time_data.coords['time'] = np.arange(nfft) / samplerate
    time_data.coords['time'].attrs['rate'] = samplerate
    return time_data


@_tools.prebind
def spectrogram(time_signal, window_duration=None, window='hann', overlap=0.5, *args, **kwargs):
    # fs = time_signal.sampling.rate
    fs = time_signal.time.rate
    window_samples = round(window_duration * fs)
    overlap_samples = round(window_duration * overlap * fs)
    f, t, Sxx = scipy.signal.spectrogram(
        x=time_signal.data,
        fs=fs,
        window=window,
        nperseg=window_samples,
        noverlap=overlap_samples,
        axis=time_signal.dims.index('time'),
    )
    dims = list(time_signal.dims)
    dims[dims.index('time')] = 'frequency'
    dims.append('time')
    return TimeFrequencyData(
        data=Sxx.copy(),  # Using a copy here is a performance improvement in later processing stages.
        # The array returned from the spectrogram function is the real part of the original stft, reshaped.
        # This means that the array takes twice the memory (the imaginary part is still around),
        # and it's not contiguous which slows down filtering a lot.
        samplerate=fs / (window_samples - overlap_samples),
        start_time=time_signal.time[0] + np.timedelta64(int(t[0] * 1e9), "ns"),
        frequency=f,
        bandwidth=fs / window_samples,
        dims=tuple(dims),
        coords=time_signal.coords,
    ).data
    return recordings.time_frequency_data(
        data=Sxx.copy(),  # Using a copy here is a performance improvement in later processing stages.
        # The array returned from the spectrogram function is the real part of the original stft, reshaped.
        # This means that the array takes twice the memory (the imaginary part is still around),
        # and it's not contiguous which slows down filtering a lot.
        samplerate=fs / (window_samples - overlap_samples),
        start_time=time_signal.sampling.window.start.add(seconds=t[0]),
        frequency=f,
        bandwidth=fs / window_samples,
        dims=tuple(dims),
        coords=time_signal.coords,
    )


@_tools.prebind
def nth_decade_filter(
    time_signal,
    bands_per_decade,
    time_step=None,
    window_duration=None,
    overlap=None,
    lower_bound=None,
    upper_bound=None,
    hybrid_resolution=False,
    scaling='density',
):
    if None not in (time_step, window_duration):
        overlap = 1 - time_step / window_duration
    elif time_step is not None:
        if overlap is None:
            if hybrid_resolution:
                # Set overlap to achieve hybrid resolution
                overlap = 1 - time_step * hybrid_resolution
            else:
                overlap = 0.5
        window_duration = time_step / (1 - overlap)
    elif window_duration is not None:
        # Cannot set overlap from hybrid resolution, the window duration is already set.
        if overlap is None:
            overlap = 0.5
        time_step = window_duration * (1 - overlap)
    elif hybrid_resolution:
        window_duration = 1 / hybrid_resolution
        if overlap is None:
            overlap = 0.5
        time_step = window_duration * (1 - overlap)
    else:
        # TODO: We could possibly use the entire time signal?
        raise ValueError('Must give at least one of `time_step` and `window_duration`.')

    if not (lower_bound or hybrid_resolution):
        raise ValueError(
            'Cannot have a log-spaced filterbank without lower frequency bound. Specify either `lower_bound` or `hybrid_resolution`.'
        )

    # We could relax these if we want to interpolate. This needs to be implemented in the calculations below.
    if hybrid_resolution:
        if hybrid_resolution is True:
            hybrid_resolution = 1 / window_duration
        if hybrid_resolution * window_duration < 1:
            raise ValueError(
                f'Hybrid filterbank with resolution of {hybrid_resolution:.2f} Hz '
                f'cannot be calculated from temporal windows of {window_duration:.2f} s.'
            )
    else:
        lowest_bandwidth = lower_bound * (10 ** (0.5 / bands_per_decade) - 10 ** (-0.5 / bands_per_decade))
        if lowest_bandwidth * window_duration < 1:
            raise ValueError(
                f'{bands_per_decade}th-decade filter band at {lower_bound:.2f} Hz with bandwidth of {lowest_bandwidth:.2f} Hz '
                f'cannot be calculated from temporal windows of {window_duration:.2f} s.'
            )

    spec = spectrogram(
        time_signal=time_signal,
        window_duration=window_duration,
        overlap=overlap,
        window=('tukey', 2 * overlap),
    ).transpose('frequency', ...)  # Put the frequency axis first for ease of indexing later

    log_band_scaling = 10 ** (0.5 / bands_per_decade)
    upper_bound = upper_bound or spec.frequency.data[-1] / log_band_scaling
    # Get frequency vectors
    if hybrid_resolution:
        minimum_bandwidth_frequency = hybrid_resolution / (log_band_scaling - 1 / log_band_scaling)
        first_log_idx = np.math.ceil(bands_per_decade * np.log10(minimum_bandwidth_frequency / 1e3))
        last_linear_idx = np.math.floor(minimum_bandwidth_frequency / hybrid_resolution)

        while (last_linear_idx + 0.5) * hybrid_resolution > 1e3 * 10 ** ((first_log_idx - 0.5) / bands_per_decade):
            # Condition is "upper edge of last linear band is higher than lower edge of first logarithmic band"
            last_linear_idx += 1
            first_log_idx += 1

        if last_linear_idx * hybrid_resolution > upper_bound:
            last_linear_idx = np.math.floor(upper_bound / hybrid_resolution)
        if lower_bound is not None:
            first_linear_idx = np.math.ceil(lower_bound / hybrid_resolution)
        else:
            first_linear_idx = 0
    else:
        first_linear_idx = last_linear_idx = 0
        first_log_idx = np.round(bands_per_decade * np.log10(lower_bound / 1e3))

    last_log_idx = round(bands_per_decade * np.log10(upper_bound / 1e3))

    lin_centers = np.arange(first_linear_idx, last_linear_idx) * hybrid_resolution
    lin_lowers = lin_centers - 0.5 * hybrid_resolution
    lin_uppers = lin_centers + 0.5 * hybrid_resolution

    log_centers = 1e3 * 10 ** (np.arange(first_log_idx, last_log_idx + 1) / bands_per_decade)
    log_lowers = log_centers / log_band_scaling
    log_uppers = log_centers * log_band_scaling

    centers = np.concatenate([lin_centers, log_centers])
    lowers = np.concatenate([lin_lowers, log_lowers])
    uppers = np.concatenate([lin_uppers, log_uppers])

    spec_data = spec.data
    banded_data = np.full(centers.shape + spec_data.shape[1:], np.nan)
    spectral_resolution = 1 / window_duration

    for idx, (l, u) in enumerate(zip(lowers, uppers)):
        l_idx = np.math.floor(l / spectral_resolution + 0.5)  # + 0.5 to consider fft bin lower edge
        u_idx = np.math.ceil(u / spectral_resolution - 0.5)  # - 0.5 to consider fft bin upper edge
        l_idx = max(l_idx, 0)
        u_idx = min(u_idx, spec_data.shape[0] - 1)

        if l_idx == u_idx:
            # This can only happen if both frequencies l and u are within the same fft bin.
            # Since we don't allow the fft bins to be larger than the output bins, we thus have the exact same band.
            banded_data[idx] = spec_data[l_idx]
        else:
            first_weight = l_idx + 0.5 - l / spectral_resolution
            last_weight = u / spectral_resolution - u_idx + 0.5
            # Sum the components fully within the output bin `[l_idx + 1:u_idx]`, and weighted components partially in the band.
            this_band = (
                spec_data[l_idx + 1 : u_idx].sum(axis=0)
                + spec_data[l_idx] * first_weight
                + spec_data[u_idx] * last_weight
            )
            banded_data[idx] = this_band * (spectral_resolution / (u - l))  # Rescale the power density.
    banded = TimeFrequencyData(
        data=banded_data,
        start_time=spec.time[0],
        samplerate=spec.time.rate,
        frequency=centers,
        bandwidth=uppers - lowers,
        dims=('frequency',) + spec.dims[1:],
        coords=spec.coords,
    ).data
    # banded = recordings.time_frequency_data(
    #     data=banded_data,
    #     start_time=spec.sampling.window.start,
    #     samplerate=spec.sampling.rate,
    #     frequency=centers,
    #     bandwidth=uppers - lowers,
    #     dims=('frequency',) + spec.dims[1:],
    #     coords=spec.coords,
    # )
    if not scaling == 'density':
        banded *= banded.bandwidth
    return banded


decidecade_filter = nth_decade_filter(bands_per_decade=10, hybrid_resolution=False)
hybrid_millidecade_filter = nth_decade_filter(bands_per_decade=1000, hybrid_resolution=1)


def convert_to_radiated_noise(source_power, source_depth, mode=None, power=True):
    if mode is None or not mode:
        return source_power
    kd = 2 * np.pi * source_power.frequency / 1500 * source_depth
    mode = mode.lower()
    if mode == 'iso':
        compensation = (14 * kd**2 + 2 * kd**4) / (14 + 2 * kd**2 + kd**4)
    elif mode == 'average farfield':
        compensation = 1 / (1 / 2 + 1 / (2 * kd**2))
    elif mode == 'isomatch':
        truncation_angle = np.radians(54.3)
        lf_comp = 2 * kd**2 * (truncation_angle - np.sin(truncation_angle) * np.cos(truncation_angle)) / truncation_angle
        compensation = 1 / (1 / 2 + 1 / lf_comp)
    elif mode == 'none':
        compensation = 1
    else:
        raise ValueError(f"Unknown mode '{mode}'")
    if power:
        return source_power * compensation
    else:
        return source_power + 10 * np.log10(compensation)


def probabilistic_spectrum(spectrogram):
    spectrogram = dB(spectrogram)
    lower_edge = np.floor(spectrogram.min())
    upper_edge = np.ceil(spectrogram.max())
    edges = np.arange(lower_edge, upper_edge + 2) - 0.5
    def hist(x):
        return np.histogram(x, bins=edges)[0]
    prob_spec = xr.apply_ufunc(
        hist,
        spectrogram,
        vectorize=True,
        # kwargs={'bins': edges},
        input_core_dims=[['time']],
        output_core_dims=[['level']],
    )
    prob_spec.coords['level'] = (edges[1:] + edges[:-1]) / 2
    return prob_spec



"""Storage and reading of recorded acoustic signals."""
import pendulum
import numpy as np
from . import positional, signals, _core
import os
import re
import soundfile
import abc


def _read_chunked_files(files, start_time, stop_time, allowable_interrupt=1):
    """Read data spread over multiple files

    Parameters
    ----------
    files : list of RecordedFile or other object with compatible API.
        List of information about the files. Each info item must have the attributes
        `samplerate`, `start_time` and `stop_time`. The list must be ordered in chronological
        order. The files must also have a `read_data(start_idx, stop_idx)` method, which
        reads the file at the designated sample indices, returning a numpy array of shape
        `(ch, stop_idx - start_idx)` or `(stop_idx - start_idx,)`. Omitting either of the
        indices should default to reading from the start of the file and to the end of the
        file respectively.
    start_time : datetime
        The start of the segment to read.
    stop_time : datetime
        The end of the segment to read.
    """

    # NOTE: We calculate the sample indices in this "collection" function and not in the file.read_data
    # functions for a reason. In many cases the start and stop times in the file labels are not perfect,
    # but the data is actually written without dropouts or repeats.
    # This means that if we allow each file to calculate it's own indices, we can end up with incorrect number
    # of read samples based only on the sporadic time labels in the files.
    # E.g. say that file 0 has a timestamp 10:00:00 and is 60 minutes and 1.5 seconds long.
    # File 1 would then have the timestamp 11:00:01, but it actually starts at 11:00:01.500.
    # Now, asking for data from 11:00:00 to 11:00:02 we expect 2*samplerate number of samples.
    # File 0 will read 1.5 seconds of data, regardless of where we calculate the sample indices.
    # Calculating the indices in the file-local functions, file 1 wold read 1 second of data.
    # Calculating the indices in the collection function, we would know that we have read 1.5 seconds
    # of data, and ask file 1 for 0.5 seconds of data.
    # This could be remedied if we update the file start times from the file stop time of the previous file,
    # but until such a procedure is implemented upon file gathering, we stick with calculating sample indices here.
    samplerate = files[0].samplerate
    samples_to_read = round((stop_time - start_time).total_seconds() * samplerate)
    for file in reversed(files):
        if file.start_time <= start_time:
            break
    else:
        raise ValueError(f'Cannot read data starting from {start_time}, earliest file start is {file.start_time}')

    if stop_time <= file.stop_time:
        # The requested data exists within one file.
        # Read the data from file and add it to the signal array.
        start_idx = np.math.floor((start_time - file.start_time).total_seconds() * samplerate)
        stop_idx = start_idx + samples_to_read
        read_signals = file.read_data(start_idx=start_idx, stop_idx=stop_idx)
    else:
        # The requested data spans multiple files
        files_to_read = []
        for file in files[files.index(file):]:
            files_to_read.append(file)
            if file.stop_time >= stop_time:
                break
        else:
            raise ValueError(f'Cannot read data extending to {stop_time}, last file ends at {file.stop_time}')

        # Check that the file boundaries are good
        for early, late in zip(files_to_read[:-1], files_to_read[1:]):
            interrupt = (late.start_time - early.stop_time).total_seconds()
            if interrupt > allowable_interrupt:
                raise ValueError(
                    f'Data is not continuous, missing {interrupt} seconds between files '
                    f'ending at {early.stop_time} and starting at {late.start_time}\n'
                    f'{early.name}\n{late.name}'
                )

        read_chunks = []

        start_idx = np.math.floor((start_time - files_to_read[0].start_time).total_seconds() * samplerate)
        chunk = files_to_read[0].read_data(start_idx=start_idx)
        read_chunks.append(chunk)
        remaining_samples = samples_to_read - chunk.shape[-1]

        for file in files_to_read[1:-1]:
            chunk = file.read_data()
            read_chunks.append(chunk)
            remaining_samples -= chunk.shape[-1]
        chunk = files_to_read[-1].read_data(stop_idx=remaining_samples)
        read_chunks.append(chunk)
        remaining_samples -= chunk.shape[-1]
        assert remaining_samples == 0

        read_signals = np.concatenate(read_chunks, axis=-1)
    return read_signals


class RecordTimeCompensation:
    """Compensates time drift and offset in a recording.

    This is based on the actual and recorded time of one or more events.
    These have to be detected elsewhere, and the times for them are
    given here to build the model.
    If a single pair of times is given, the offset between them is used to compensate.
    If multiple pairs are given, the offset will be linearly interpolated between them.

    Parameters
    ----------
    actual_time : time_like or [time_like]
        Actual time for synchronization event(s).
    recorded_time : time_like or [time_like]
        Recorded time for synchronization event(s).
    """
    def __init__(self, actual_time, recorded_time):
        if isinstance(actual_time, str):
            actual_time = [actual_time]
        if isinstance(recorded_time, str):
            recorded_time = [recorded_time]
        try:
            iter(actual_time)
        except TypeError:
            actual_time = [actual_time]
        try:
            iter(recorded_time)
        except TypeError:
            recorded_time = [recorded_time]

        actual_time = list(map(positional._sanitize_datetime_input, actual_time))
        recorded_time = list(map(positional._sanitize_datetime_input, recorded_time))

        self._time_offset = [(recorded - actual).total_seconds() for (recorded, actual) in zip(recorded_time, actual_time)]
        if len(self._time_offset) > 1:
            self._actual_timestamps = [t.timestamp() for t in actual_time]
            self._recorded_timestamps = [t.timestamp() for t in recorded_time]

    def recorded_to_actual(self, recorded_time):
        recorded_time = positional._sanitize_datetime_input(recorded_time)
        if len(self._time_offset) == 1:
            time_offset = self._time_offset[0]
        else:
            time_offset = np.interp(recorded_time.timestamp(), self._recorded_timestamps, self._time_offset)
        return recorded_time - pendulum.duration(seconds=time_offset)

    def actual_to_recorded(self, actual_time):
        actual_time = positional._sanitize_datetime_input(actual_time)
        if len(self._time_offset) == 1:
            time_offset = self._time_offset[0]
        else:
            time_offset = np.interp(actual_time.timestamp(), self._actual_timestamps, self._time_offset)
        return actual_time + pendulum.duration(seconds=time_offset)


class RecordedFile(abc.ABC):
    def __init__(self, name):
        self.name = name

    @abc.abstractmethod
    def read_info(self):
        ...

    @abc.abstractmethod
    def read_data(self, start_idx, stop_idx):
        ...

    @staticmethod
    def _lazy_property(key):
        def getter(self):
            try:
                return getattr(self, '_' + key)
            except AttributeError:
                self.read_info()
            return getattr(self, '_' + key)
        return property(getter)

    @property
    @abc.abstractmethod
    def samplerate(self):
        ...

    @property
    @abc.abstractmethod
    def start_time(self):
        ...

    @property
    @abc.abstractmethod
    def stop_time(self):
        ...

    @property
    def duration(self):
        return (self.stop_time - self.start_time).total_seconds()

    def __bool__(self):
        return os.path.exists(self.name)


class Recording(_core.TimeLeafin, _core.Leaf):
    @property
    @abc.abstractmethod
    def data(self):
        ...


class Hydrophone(Recording):
    def __init__(
        self,
        calibration,
        channel=None,
        position=None,
        depth=None,
        metadata=None,
        **kwargs
    ):
        metadata = metadata or {}
        if channel is not None:
            metadata['channel'] = channel
        if position is not None:
            metadata['hydrophone position'] = positional.Position(position)
        if depth is not None:
            metadata['hydrophone depth'] = depth
        super().__init__(**kwargs, metadata=metadata)
        self.calibration = calibration

    @property
    def time_period(self):
        return self._time_period


class SplitFileHydrophone(Hydrophone):
    allowable_interrupt = 1

    def __init__(self, files, **kwargs):
        super().__init__(**kwargs)
        self.files = files
        start_time = self.files[0].start_time
        stop_time = self.files[-1].stop_time
        self._time_period = _core.TimePeriod(start=start_time, stop=stop_time)

    @property
    def data(self):
        read_signals = _read_chunked_files(
            files=self.files,
            start_time=self.time_period.start,
            stop_time=self.time_period.stop,
            allowable_interrupt=self.allowable_interrupt,
        )

        if self.calibration is None:
            signal = signals.Time(
                data=read_signals,
                samplerate=self.samplerate,
                start_time=self.time_period.start,
                metadata=self.metadata.data
            )
        else:
            signal = signals.Pressure.from_raw_and_calibration(
                data=read_signals,
                calibration=self.calibration,
                samplerate=self.samplerate,
                start_time=self.time_period.start,
                metadata=self.metadata.data
            )

        return signal

    def time_subperiod(self, time=None, /, *, start=None, stop=None, center=None, duration=None):
        time_period = self.time_period.subperiod(time, start=start, stop=stop, center=center, duration=duration)
        obj = self.clone()
        obj._time_period = time_period
        return obj

    @property
    def samplerate(self):
        return self.files[0].samplerate

    def clone(self):
        return type(self)(self.files, metadata=self.metadata, calibration=self.calibration)


def sound_trap(folder, serial_number, time_compensation=None, **kwargs):
    if time_compensation is None:
        def time_compensation(timestamp):
            return timestamp
    if isinstance(time_compensation, RecordTimeCompensation):
        time_compensation = time_compensation.recorded_to_actual
    elif not callable(time_compensation):
        offset = pendulum.duration(seconds=time_compensation)
        def time_compensation(timestamp):
            return timestamp - offset

    files = []
    for file in sorted(filter(lambda x: x.is_file(), os.scandir(folder)), key=lambda x: x.name):
        file = SoundTrap.RecordedFile(name=os.path.join(folder, file), time_compensation=time_compensation)
        if file and (file.serial_number == serial_number):
            files.append(file)

    return SplitFileHydrophone(files, channel=serial_number, **kwargs)


class HydrophoneArray(_core.TimeBranchin, _core.Branch):
    def __init__(self, hydrophones, position=None, depth=None, metadata=None):
        metadata = metadata or {}
        if position is not None:
            metadata['hydrophone position'] = positional.Position(position)
        if depth is not None:
            metadata['hydrophone depth'] = depth
        super().__init__(dim='channel', children=hydrophones, metadata=metadata)

    @property
    def hydrophones(self):
        return self._children

    @property
    def data(self):
        return signals.DataStack(
            dim=self.dim,
            children={name: hydrophone.data for name, hydrophone in self.items()},
            metadata=self.metadata.data,
        )

    def clone(self, hydrophones):
        return type(self)(hydrophones, metadata=self.metadata)


class SoundTrap(Hydrophone):
    #  The file starts are taken from the timestamps in the filename, which is quantized to 1s.
    allowable_interrupt = 1

    class RecordedFile(RecordedFile):
        pattern = r'(\d{4})\.(\d{12}).wav'
        def __init__(self, name, time_compensation):
            super().__init__(name=name)
            name = os.path.basename(name)
            if not (match := re.match(self.pattern, name)):
                return
            serial_number, time = match.groups()
            self.serial_number = int(serial_number)
            self._start_time = time_compensation(pendulum.from_format(time, 'YYMMDDHHmmss'))

        def __bool__(self):
            return super().__bool__() and hasattr(self, 'serial_number') and bool(self.serial_number)

        def read_info(self):
            sfi = soundfile.info(self.name)
            self._stop_time = self.start_time + pendulum.duration(seconds=sfi.duration)
            self._samplerate = sfi.samplerate

        start_time = RecordedFile._lazy_property('start_time')
        stop_time = RecordedFile._lazy_property('stop_time')
        samplerate = RecordedFile._lazy_property('samplerate')

        def read_data(self, start_idx=None, stop_idx=None):
            return soundfile.read(self.name, start=start_idx, stop=stop_idx, dtype='float32')[0]

    def __init__(self, folder, serial_number, time_compensation=None, calibration=None, **kwargs):
        """Read a folder with SoundTrap data.

        Parameters
        ----------
        folder : str
            Path to the folder with the data.
        key : int, str
            The serial number of the Hydrophone.
            Can be given as an integer or a string.
        calibrations : dict or numeric
            A dict with the calibration values of the SoundTraps.
            If a single value is given, it will be used for all read data.
            Give as a value in dB re. 1/μPa, e.g. -188.5
        depth : dict or numeric
            A dict with the depths of the SoundTraps, in meters.
            If a single value is given, it will be used for all read data.
        time_offset : {numeric, callable}, optional
            Time offset that will be added to the timestamps stored in files.
            For custom offsets, pass a function with the signature
                `offset = time_offset(timestamp, serial_number)`
            that returns the offset for the file timestamp and particular serial number.
        """
        super().__init__(channel=serial_number, **kwargs)
        self.calibration = calibration
        self.folder = folder

        if time_compensation is None:
            def time_compensation(timestamp):
                return timestamp
        if isinstance(time_compensation, RecordTimeCompensation):
            time_compensation = time_compensation.recorded_to_actual
        elif not callable(time_compensation):
            offset = pendulum.duration(seconds=time_compensation)
            def time_compensation(timestamp):
                return timestamp - offset

        self.files = []
        for file in sorted(filter(lambda x: x.is_file(), os.scandir(self.folder)), key=lambda x: x.name):
            file = self.RecordedFile(name=os.path.join(self.folder, file), time_compensation=time_compensation)
            if file and (file.serial_number == self.serial_number):
                self.files.append(file)
        start_time = self.files[0].start_time
        stop_time = self.files[-1].stop_time
        self._time_period = _core.TimePeriod(start=start_time, stop=stop_time)

    @property
    def samplerate(self):
        return self.files[0].samplerate

    @property
    def serial_number(self):
        return int(self.metadata['channel'])

    def copy(self, **kwargs):
        obj = super().copy(**kwargs)
        obj.folder = self.folder
        obj.files = self.files
        obj.calibration = self.calibration
        return obj

    @property
    def data(self):
        read_signals = _read_chunked_files(
            files=self.files,
            start_time=self.time_period.start,
            stop_time=self.time_period.stop,
            allowable_interrupt=self.allowable_interrupt,
        )

        if self.calibration is None:
            signal = signals.Time(
                data=read_signals,
                samplerate=self.samplerate,
                start_time=self.time_period.start,
                metadata=self.metadata.data
            )
        else:
            signal = signals.Pressure.from_raw_and_calibration(
                data=read_signals,
                calibration=self.calibration,
                samplerate=self.samplerate,
                start_time=self.time_period.start,
                metadata=self.metadata.data
            )

        return signal


class SylenceLP(Hydrophone):
    allowable_interrupt = 1
    voltage_range = 2.5

    class RecordedFile(RecordedFile):
        patten = r"channel([A-D])_(\d{4})-(\d{2})-(\d{2})_(\d{2})-(\d{2})-(\d{2}).wav"
        def __init__(self, name, time_compensation):
            super().__init__(name)
            if not super().__bool__():
                return

            basename = os.path.basename(self.name)
            if not (match := re.match(self.patten, basename)):
                return
            channel, year, month, day, hour, minute, second = match.groups()
            self._start_time = time_compensation(pendulum.datetime(
                int(year), int(month), int(day),
                int(hour), int(minute), int(second),
            ))

        def __bool__(self):
            return super().__bool__() and hasattr(self, '_start_time')

        def read_data(self, start_idx=None, stop_idx=None):
            return soundfile.read(self.name, start=start_idx, stop=stop_idx, dtype='float32')[0]

        def read_info(self):
            with open(self.name, 'rb') as file:
                base_header = file.read(36)
                # chunk_id = base_header[0:4].decode('ascii')  # always equals RIFF
                # file_size = int.from_bytes(base_header[4:8], byteorder='little', signed=False)  # total file size not important
                # chunk_format = base_header[8:12].decode('ascii')  # always equals WAVE
                # subchunk_id = base_header[12:16].decode('ascii')  # always equals fmt
                # subchunk_size = int.from_bytes(base_header[16:20], byteorder='little', signed=False))  # always equals 16
                # audio_format = int.from_bytes(base_header[20:22], byteorder='little', signed=False))  # not important in current implementation
                num_channels = int.from_bytes(base_header[22:24], byteorder='little', signed=False)
                if num_channels != 1:
                    raise ValueError(f"Expected file for SylenceLP with a single channel, read file with {num_channels} channels")
                samplerate = int.from_bytes(base_header[24:28], byteorder='little', signed=False)
                # byte rate = int.from_bytes(base_header[28:32], byteorder='little', signed=False)  # not important in current implementation
                bytes_per_sample = int.from_bytes(base_header[32:34], byteorder='little', signed=False)
                bitdepth = int.from_bytes(base_header[34:36], byteorder='little', signed=False)

                conf_header = file.peek(8)  # uses peak to keep indices aligned with the manual
                conf_size = int.from_bytes(conf_header[4:8], byteorder='little', signed=False)
                if conf_size != 460:
                    raise ValueError(f"Incorrect size of SylenceLP config: '{conf_size}'B, expected 460B")
                conf_header = file.read(conf_size + 8)

                subchunk_id = conf_header[:4].decode('ascii')  # always conf
                if subchunk_id != 'conf':
                    raise ValueError(f"Expected 'conf' section in SylenceLP config, found '{subchunk_id}'")
                # subchunk_size = int.from_bytes(conf_header[4:8], byteorder='little', signed=False)  # the same as conf_size
                config_version = int.from_bytes(conf_header[8:12], byteorder='little', signed=False)
                if config_version != 2:
                    raise NotImplementedError(f'Cannot handle SylenceLP config version {config_version}')
                # recording_start = datetime.datetime.fromtimestamp(int.from_bytes(conf_header[16:24], byteorder='little', signed=True))  # This value is not actually when the recording starts. No idea what it actually is
                channel = conf_header[24:28].decode('ascii')
                if channel.strip('\x00') != '':
                    raise NotImplementedError(f"No implementation for multichannel SylenceLP recorders, found channel specification '{channel}'")
                samplerate_alt = np.frombuffer(conf_header[28:32], dtype='f4').squeeze()
                if samplerate != samplerate_alt:
                    raise ValueError(f"Mismatched samplerate for hardware and file, read file samplerate {samplerate} and config samplerate {samplerate_alt}")

                hydrophone_sensitivity = np.frombuffer(conf_header[32:48], dtype='f4')
                gain = np.frombuffer(conf_header[48:64], dtype='f4')
                # gain_correction = np.frombuffer(conf_header[64:80], dtype='f4')  # is just 1/gain
                serialnumber = conf_header[80:100].decode('ascii')
                active_channels = conf_header[100:104].decode('ascii')
                if active_channels != 'A\x00\x00\x00':
                    raise NotImplementedError(f"No implementation for multichannel SylenceLP recorders, found channel specification '{active_channels}'")

                data_header = file.read(4).decode('ascii')
                if data_header != 'data':
                    raise ValueError(f"Expected file header 'data', read {data_header}")
                data_size = int.from_bytes(file.read(4), byteorder='little', signed=False)

            num_samples = data_size / bytes_per_sample
            if int(num_samples) != num_samples:
                raise ValueError(f"Size of data is not divisible by bytes per sample, file '{self.name}' is corrupt!")

            self._samplerate = samplerate
            self._bitdepth = bitdepth
            # self._start_time = recording_start  # The start property in the file headers is incorrect... It might be the timestamp when the file was created, but in local time instead of UTC? This is useless since the files are pre-created.
            self._stop_time = self.start_time + pendulum.duraion(seconds=num_samples / samplerate)
            self._hydrophone_sensitivity = hydrophone_sensitivity[0]
            self._serial_number = serialnumber.strip('\x00')
            self._gain = 20 * np.log10(gain[0])

        samplerate = RecordedFile._lazy_property('samplerate')
        bitdepth = RecordedFile._lazy_property('bitdepth')
        start_time = RecordedFile._lazy_property('start_time')
        stop_time = RecordedFile._lazy_property('stop_time')
        hydrophone_sensitivity = RecordedFile._lazy_property('hydrophone_sensitivity')
        serial_number = RecordedFile._lazy_property('serial_number')
        gain = RecordedFile._lazy_property('gain')

    def __init__(self, folder, time_compensation=None, **kwargs):
        super().__init__(**kwargs)
        self.folder = folder
        self.files = []

        if time_compensation is None:
            def time_compensation(timestamp):
                return timestamp
        elif isinstance(time_compensation, RecordTimeCompensation):
            time_compensation = time_compensation.recorded_to_actual
        elif not callable(time_compensation):
            offset = pendulum.duration(seconds=time_compensation)
            def time_compensation(timestamp):
                return timestamp - offset

        for directory in sorted(filter(lambda x: x.is_dir(), os.scandir(self.folder)), key=lambda x: x.name):
            for file in sorted(filter(lambda x: x.is_file(), os.scandir(directory.path)), key=lambda x: x.name):
                if file := self.RecordedFile(file.path, time_compensation=time_compensation):
                    self.files.append(file)

        start_time = self.files[0].start_time
        stop_time = self.files[-1].stop_time
        self._time_period = _core.TimePeriod(start=start_time, stop=stop_time)

    @property
    def samplerate(self):
        return self.files[0].samplerate

    @property
    def serial_number(self):
        return self.files[0].serial_number

    @property
    def calibration(self):
        hydrophone_sensitivity = self.files[0].hydrophone_sensitivity
        gain = self.files[0].gain
        return hydrophone_sensitivity + gain - 20 * np.log10(self.voltage_range)

    def copy(self, **kwargs):
        obj = super().copy(**kwargs)
        obj.folder = self.folder
        obj.files = self.files
        return obj

    @property
    def data(self):
        read_signals = _read_chunked_files(
            files=self.files,
            start_time=self.time_period.start,
            stop_time=self.time_period.stop,
            allowable_interrupt=self.allowable_interrupt,
        )

        signal = signals.Pressure.from_raw_and_calibration(
            data=read_signals,
            calibration=self.calibration,
            samplerate=self.samplerate,
            start_time=self.time_period.start,
            metadata=self.metadata.data
        )
        return signal


class MultichannelAudioPassage(Hydrophone):
    class _Sampling(Hydrophone._Sampling):
        @property
        def rate(self):
            return self.hydrophone.audio_info.samplerate

        @property
        def window(self):
            try:
                return self._window
            except AttributeError:
                self._window = positional.TimeWindow(
                    start=self.hydrophone.audio_info.start_time,
                    duration=self.hydrophone.audio_info.duration,
                )
                return self._window

        def subwindow(self, time=None, /, *, start=None, stop=None, center=None, duration=None):
            original_window = self.window
            new_window = original_window.subwindow(time, start=start, stop=stop, center=center, duration=duration)
            new = type(self.hydrophone)(
                filename=self.hydrophone.filename,s
                calibration=self.hydrophone.calibration,
                depth=self.hydrophone.depth,
                position=self.hydrophone.position,
                start_time=self.hydrophone.audio_info.start_time,
                channels=self.hydrophone._channels,
            )
            new.sampling._window = new_window
            return new

    def __init__(self, filename, start_time, channels=None, calibration=None, **kwargs):
        super().__init__(**kwargs)
        self.filename = filename
        self.audio_info = soundfile.info(self.filename)
        self.audio_info.start_time = positional._sanitize_datetime_input(start_time)
        self._channels = channels or list(range(self.audio_info.channels))
        self.calibration = calibration
        self.depth = xr.DataArray(self.depth, dims='receiver')

    @property
    def num_channels(self):
        return len(self._channels)

    @property
    def time_data(self):
        time_window = self.sampling.window
        start_sample = round((time_window.start - self.audio_info.start_time).total_seconds() * self.sampling.rate)
        stop_sample = round((time_window.stop - self.audio_info.start_time).total_seconds() * self.sampling.rate)
        data, fs = soundfile.read(
            self.filename,
            start=start_sample,
            stop=stop_sample,
        )
        if data.ndim == 2:
            data = data[:, self._channels]
        if data.ndim == 1:
            dims = ('time',)
        else:
            dims = ('time', 'receiver')
        return time_data(data, start_time=time_window.start, samplerate=self.sampling.rate, calibration=self.calibration, dims=dims)



from . import positional

import xarray as xr
import numpy as np


def sensor(label, sensitivity=None, position=None, depth=None):
    data_vars = {}
    if sensitivity is not None:
        data_vars['sensitivity'] = sensitivity
    if position is not None:
        position = positional.position(position)
        data_vars['latitude'] = position.latitude
        data_vars['longitude'] = position.longitude
    if depth is not None:
        data_vars['depth'] = depth

    return xr.Dataset(data_vars=data_vars, coords={'sensor': label})


def sensor_array(*sensors, **kwargs):
    """Collects sensor information in an xarray.

    The function accepts two types of calls, positional sensors or keywords with dicts.
    The positional format is `sensor_array(sensor_1, sensor_2, ...)`
    where each sensor is an xarray with at least `sensor` containing the label for that sensor.
    Such xarrays can be created with the `sensor` function.
    The other format is keyword arguments with said labels as the keys, and a dictionary
    with the sensor information as the value, e.g.,

    ```
    sensor_array(
        soundtrap_1={'position': (58.25, 11.14), 'sensitivity': -182},
        soundtrap_2={'position': (58.26, 11.15), 'sensitivity': -183},
    )
    ```
    Note that labels that are not valid arguments can still be created using dict unpacking
    ```
    sensor_array(**{
        'SoundTrap 1': {'position': (58.25, 11.14), 'sensitivity': -182},
        'SoundTrap 2': {'position': (58.26, 11.15), 'sensitivity': -183},
    })
    """
    if kwargs:
        sensors = sensors + (
            sensor(label, **values)
            for label, values in kwargs.items()
        )
    sensors = xr.concat(sensors, dim='sensor')
    for key, value in sensors.items():
        if np.ptp(value.values) == 0:
            sensors[key] = value.mean()
    return sensors


def align_property_to_sensors(sensors, values, allow_scalar=False):
    sensor_names = sensors.sensor

    if isinstance(values, xr.DataArray):
        return values
    if allow_scalar:
        try:
            len(values)
        except TypeError:
            return values

    if len(values) != sensor_names.size:
        raise ValueError(f"Cannot assign {len(values)} values to {sensor_names.size} sensors")

    try:
        return xr.DataArray(values, coords={'sensor': sensor_names})
    except ValueError:
        pass
    return xr.DataArray([values[key] for key in sensor_names.values], coords={'sensor': sensor_names})



import plotly.graph_objects as go
import numpy as np


def make_chart(center, zoom=14, mapbox_accesstoken=None):
    if mapbox_accesstoken is None:
        import os
        mapbox_accesstoken = os.getenv("MAPBOX_ACCESSTOKEN")
        if mapbox_accesstoken is None:
            import dotenv
            mapbox_accesstoken = dotenv.get_key(dotenv.find_dotenv(), "MAPBOX_ACCESSTOKEN")
    return go.Figure(layout=dict(
        mapbox=dict(
            accesstoken=mapbox_accesstoken,
            style='light',
            zoom=zoom,
            center=dict(lat=float(center.latitude), lon=float(center.longitude)),
        ),
        height=1200,
        margin={"r":0,"t":0,"l":0,"b":0},
        legend=dict(
            orientation="h",
            yanchor="bottom",
            y=1.02,
            xanchor="left",
            x=0,
        )
    ))


def plot_track(track, **kwargs):
    kwargs['lat'] = np.atleast_1d(track.latitude)
    kwargs['lon'] = np.atleast_1d(track.longitude)
    if hasattr(track, 'time'):
        if (
            track.time.size == track.latitude.size
            and 'hovertemplate' not in kwargs
            and 'customdata' not in kwargs
        ):
            kwargs['customdata'] = track.time
            kwargs['hovertemplate'] = '(%{lat}º, %{lon}º), %{customdata|%X}'
        else:
            kwargs.setdefault('hovertemplate', '(%{lat}º, %{lon}º)')
    return go.Scattermapbox(**kwargs)


def listen(recording, downsampling=1, upsampling=None, **kwargs):
    import sounddevice as sd
    from .recordings import Recording
    sd.stop()
    if isinstance(recording, Recording):
        recording = recording.time_data()
    if upsampling:
        recording = recording[::upsampling]
    scaled = recording - recording.mean()
    scaled = scaled / np.max(np.abs(scaled))
    sd.play(scaled, samplerate=round(recording.sampling.rate / downsampling), **kwargs)



# %%
import uwacan
geod = uwacan.positional.geod
import numpy as np
import plotly.graph_objects as go
import folium
5

from importlib import reload
reload(uwacan.positional)
hydrophone_position = uwacan.positional.Position.from_degrees_minutes_seconds(57, 35.424, 11, 48.350)
# %%
import gpxpy
path = 'c:/Users/carl4189/OneDrive - IVL Svenska Miljöinstitutet AB/Elvy maj 22/GPS-data/GPS1_dag.gpx'
file = open(path, 'r')
contents = gpxpy.parse(file)
latitudes = []
longitudes = []
times = []
for point in contents.get_points_data():
    latitudes.append(point.point.latitude)
    longitudes.append(point.point.longitude)
    times.append(point.point.time)

# %%
track = uwacan.positional.GPXTrack('c:/Users/carl4189/OneDrive - IVL Svenska Miljöinstitutet AB/Elvy maj 22/GPS-data/GPS1_dag.gpx')

# %%
track = uwacan.positional.Blueflow('c:/Users/carl4189/OneDrive - IVL Svenska Miljöinstitutet AB/Elvy maj 22/Blueflow/Blueflow data Elvy 17 maj.xlsx')

# %%

m = folium.Map(location=[np.mean(track.latitude), np.mean(track.longitude)], zoom_start=13, control_scale=True)
folium.PolyLine(
    np.stack([track.latitude, track.longitude], axis=1),
    weight=1,
    lineCap='bevel',
    color='red'
).add_to(m)
display.display(m)

# %%
subtrack = track[40:50]
revtrack = reversed(subtrack)
list(revtrack)
# %%
start_idx = 3100
track_len = 400
stop_idx = start_idx + track_len
subtrack = track[start_idx:stop_idx]
# latitude = track.latitude[start_idx:stop_idx]
# longitude = track.longitude[start_idx:stop_idx]
# pos = np.stack([subtrack.latitude, subtrack.longitude], axis=1)
# times = [(t - track.time.timestamps[start_idx]).total_seconds() for t in track.time.timestamps[start_idx:stop_idx]]
times = subtrack.track_time

time_windows = subtrack.aspect_windows(hydrophone_position, 5, 45, 100, None)

m = folium.Map(location=subtrack.mean.coordinates, zoom_start=15, control_scale=True)
folium.PolyLine(
    subtrack.coordinates.transpose(),
    weight=1,
    lineCap='bevel',
    color='red'
).add_to(m)
folium.Marker(hydrophone_position.coordinates, popup='hydrophone').add_to(m)
for idx, window in enumerate(time_windows):
    start_pos = subtrack[window.start]
    stop_pos = subtrack[window.stop]
    print(f'{idx = }: {hydrophone_position.angle_between(start_pos, stop_pos)} degrees, {start_pos.distance_to(stop_pos)} meters')
    folium.Marker(start_pos.coordinates, icon=folium.Icon(color='green', icon='marker'), tooltip=f'{idx}', popup=str(start_pos)).add_to(m)
    folium.Marker(stop_pos.coordinates, icon=folium.Icon(color='red', icon='marker'), tooltip=f'{idx}', popup=str(stop_pos)).add_to(m)
folium.Marker(subtrack.closest_point(hydrophone_position).coordinates, icon=folium.Icon(color='blue', icon='marker'), tooltip=f'cpa').add_to(m)
display.display(m)

# %%
subtrack[window.start]
# %%
blueflow_heading = track.data['Heading [deg]'].iloc[start_idx:stop_idx].to_numpy()
# blueflow_speed = uwacan.positional.speed_conversion(knots=track.blueflow['Speed over ground [kts]'].iloc[start_idx:stop_idx].to_numpy())
blueflow_speed = track.data['Speed over ground [kts]'].iloc[start_idx:stop_idx].to_numpy() / uwacan.positional.one_knot
headings = []
speeds = []
for (early_lat, early_lon), (late_lat, late_lon), early_t, late_t in zip(pos[:-1], pos[1:], times[:-1], times[1:]):
    outs = geod.Inverse(early_lat, early_lon, late_lat, late_lon, outmask=geod.AZIMUTH | geod.DISTANCE)
    headings.append(outs['azi2'])
    speeds.append(outs['s12'] / (late_t - early_t))

go.Figure([go.Scatter(y=blueflow_heading), go.Scatter(y=headings)]).show()
go.Figure([go.Scatter(y=blueflow_speed), go.Scatter(y=speeds)]).show()

# %%

duration = 260
sampling_interval = 2
jitter = 1
n_samples = round(duration / sampling_interval)
time = np.arange(n_samples) * sampling_interval + np.random.normal(size=n_samples, scale=jitter)
time -= time[0]
time = np.insert(time, 4, time[4])

latitude = np.random.normal(size=time.shape, scale=0.5) + np.linspace(-1, 1, time.size) ** 2 + 58
longitude = np.random.normal(size=time.shape, scale=0.5) + np.linspace(11.4, 12, time.size)

# %%
track = uwacan.positional.Track(latitude=latitude, longitude=longitude, time=time)
# %%
track.resample(0.25).smooth(20)
# %%
go.Figure([
    go.Scatter(x=track.time, y=track.longitude, mode='lines'),
    go.Scatter(x=track.resample.time, y=track.resample.longitude, mode='lines'),
    go.Scatter(x=track.resample.smooth.time, y=track.resample.smooth.longitude, mode='lines'),
])
# %%



import numpy as np
import xarray as xr
from . import positional


class _DataWrapper:
    @classmethod
    def _wrap_output(cls, data):
        if not isinstance(data, xr.DataArray):
            return data
        if 'time' in data.dims and 'frequency' in data.dims:
            return TimeFrequencyData(data)
        if 'time' in data.dims:
            return TimeData(data)
        if 'frequency' in data.dims:
            return FrequencyData(data)
        return data

    def __add__(self, other):
        if isinstance(other, _DataWrapper):
            other = other.data
        return self._wrap_output(self.data + other)

    def __radd__(self, other):
        if isinstance(other, _DataWrapper):
            other = other.data
        return self._wrap_output(other + self.data)

    def __sub__(self, other):
        if isinstance(other, _DataWrapper):
            other = other.data
        return self._wrap_output(self.data - other)

    def __rsub__(self, other):
        if isinstance(other, _DataWrapper):
            other = other.data
        return self._wrap_output(other - self.data)

    def __mul__(self, other):
        if isinstance(other, _DataWrapper):
            other = other.data
        return self._wrap_output(self.data * other)

    def __rmul__(self, other):
        if isinstance(other, _DataWrapper):
            other = other.data
        return self._wrap_output(other * self.data)

    def __truediv__(self, other):
        if isinstance(other, _DataWrapper):
            other = other.data
        return self._wrap_output(self.data / other)

    def __rtruediv__(self, other):
        if isinstance(other, _DataWrapper):
            other = other.data
        return self._wrap_output(other / self.data)

    def __floordiv__(self, other):
        if isinstance(other, _DataWrapper):
            other = other.data
        return self._wrap_output(self.data // other)

    def __rfloordiv__(self, other):
        if isinstance(other, _DataWrapper):
            other = other.data
        return self._wrap_output(other // self.data)

    def __pow__(self, other):
        if isinstance(other, _DataWrapper):
            other = other.data
        return self._wrap_output(self.data ** other)

    def __rpow__(self, other):
        if isinstance(other, _DataWrapper):
            other = other.data
        return self._wrap_output(other ** self.data)

    def __mod__(self, other):
        if isinstance(other, _DataWrapper):
            other = other.data
        return self._wrap_output(self.data % other)

    def __rmod__(self, other):
        if isinstance(other, _DataWrapper):
            other = other.data
        return self._wrap_output(other % self.data)

    def __neg__(self):
        return self._wrap_output(-self.data)

    def __abs__(self):
        return self._wrap_output(abs(self.data))


def _assign_time_vector(data, start_time, samplerate):
    if samplerate is None:
        return data
    if start_time is None:
        if 'time' in data.coords:
            start_time = data.time[0].item()
        start_time = 'now'
    n_samples = data.sizes['time']
    start_time = positional.time_to_np(start_time)
    offsets = np.arange(n_samples) * 1e9 / samplerate
    time = start_time + offsets.astype('timedelta64[ns]')
    return data.assign_coords(time=('time', time, {'rate': samplerate}))


class TimeData(_DataWrapper):
    def __init__(self, data, start_time=None, samplerate=None, dims=None, coords=None):
        if not isinstance(data, xr.DataArray):
            if dims is None:
                if data.ndim == 1:
                    dims = 'time'
                else:
                    raise ValueError(f'Cannot guess dimensions for time data with {data.ndim} dimensions')
            data = xr.DataArray(data, dims=dims)

        data = data.assign_coords(coords)
        data = _assign_time_vector(data, samplerate=samplerate, start_time=start_time)
        self.data = data

    @property
    def samplerate(self):
        return self.data.time.rate

    @property
    def time_window(self):
        # Calculating duration from number and rate means the stop points to the sample after the last,
        # which is more intuitive when considering signal durations etc.
        return positional.TimeWindow(
            start=self.data.time.data[0],
            duration=self.data.sizes['time'] / self.samplerate,
        )

    def subwindow(self, time=None, /, *, start=None, stop=None, center=None, duration=None, extend=None):
        original_window = self.time_window
        new_window = original_window.subwindow(time, start=start, stop=stop, center=center, duration=duration, extend=extend)
        if isinstance(new_window, positional.TimeWindow):
            start = (new_window.start - original_window.start).total_seconds()
            stop = (new_window.stop - original_window.start).total_seconds()
            # Indices assumed to be seconds from start
            start = np.math.floor(start * self.samplerate)
            stop = np.math.ceil(stop * self.samplerate)
            idx = slice(start, stop)
        else:
            idx = (new_window - original_window.start).total_seconds()
            idx = round(idx * self.samplerate)

        selected_data = self.data.isel(time=idx)
        return type(self)(selected_data)


def _assign_frequency_bandwidth_vectors(data, frequency, bandwidth):
    if frequency is None:
        return data
    coords = {'frequency': frequency}
    if bandwidth is not None:
        coords['bandwidth'] = ('frequency', np.broadcast_to(bandwidth, np.shape(frequency)))
    return data.assign_coords(coords)


class FrequencyData(_DataWrapper):
    def __init__(self, data, frequency=None, bandwidth=None, dims=None, coords=None, **kwargs):
        if not isinstance(data, xr.DataArray):
            if dims is None:
                if data.ndim == 1:
                    dims = 'frequency'
                else:
                    raise ValueError(f'Cannot guess dimensions for frequency data with {data.ndim} dimensions')
            data = xr.DataArray(data, dims=dims)
        data = data.assign_coords(coords)
        data = _assign_frequency_bandwidth_vectors(data, frequency=frequency, bandwidth=bandwidth)
        self.data = data

    def estimate_bandwidth(self):
        frequency = np.asarray(self.data.frequency)
        # Check if the frequency array seems linearly or logarithmically spaced
        if frequency[0] == 0:
            diff = frequency[2:] - frequency[1:-1]
            frac = frequency[2:] / frequency[1:-1]
        else:
            diff = frequency[1:] - frequency[:-1]
            frac = frequency[1:] / frequency[:-1]
        diff_err = np.std(diff) / np.mean(diff)
        frac_err = np.std(frac) / np.mean(frac)
        # Note: if there are three values and the first is zero, the std is 0 for both.
        # The equals option makes us end up in the linear frequency case.
        if diff_err <= frac_err:
            # Central differences, with forwards and backwards at the ends
            central = (frequency[2:] - frequency[:-2]) / 2
            first = frequency[1] - frequency[0]
            last = frequency[-1] - frequency[-2]
        else:
            # upper edge is at sqrt(f_{l+1} * f_l), lower edge is at sqrt(f_{l-1} * f_l)
            # the difference simplifies as below.
            central = (frequency[2:]**0.5 - frequency[:-2]**0.5) * frequency[1:-1]**0.5
            # extrapolating to one bin below lowest and one above highest using constant ratio
            # the expression above then simplifies to the expressions below
            first = (frequency[1] - frequency[0]) * (frequency[0] / frequency[1])**0.5
            last = (frequency[-1] - frequency[-2]) * (frequency[-1] / frequency[-2])**0.5
        bandwidth = np.concatenate([[first], central, [last]])
        return xr.DataArray(bandwidth, coords={'frequency': self.data.frequency})


class TimeFrequencyData(TimeData, FrequencyData):
    def __init__(self, data, start_time=None, samplerate=None, frequency=None, bandwidth=None, dims=None, coords=None, **kwargs):
        if not isinstance(data, xr.DataArray):
            if dims is None:
                raise ValueError('Cannot guess dimensions for time-frequency data')
            data = xr.DataArray(data, dims=dims)
        data = data.assign_coords(coords)
        data = _assign_time_vector(data, samplerate=samplerate, start_time=start_time)
        data = _assign_frequency_bandwidth_vectors(data, frequency=frequency, bandwidth=bandwidth)
        self.data = data



import numpy as np
import scipy.signal
from . import timestamps
from . import signals


class NthOctavebandFilterBank:
    def __init__(self, frequency_range, bands_per_octave=3, filter_order=8):
        self.frequency_range = frequency_range
        self.bands_per_octave = bands_per_octave
        self.filter_order = filter_order

    def __call__(self, signal):
        if isinstance(signal, signals.Spectrogram):
            powers = self.power_filters(signal.frequencies).dot(signal.signal) / signal.frequencies[1]
            return signals.Signal(
                signal=powers,
                samplerate=signal.times.samplerate,
                start_time=signal.times.start_time,
            )
        else:
            raise TypeError(f'Cannot filter data of input type {type(signal)}')

    @property
    def center_frequencies(self):
        lowest_band, highest_band = self.frequency_range
        lowest_band_index = np.round(self.bands_per_octave * np.log2(lowest_band / 1e3))
        highest_band_index = np.round(self.bands_per_octave * np.log2(highest_band / 1e3))
        octaves = np.arange(lowest_band_index, highest_band_index + 1) / self.bands_per_octave
        return 1e3 * 2 ** octaves

    def power_filters(self, frequencies):
        centers = self.center_frequencies[:, None]
        bandwidths = centers * (2**(0.5 / self.bands_per_octave) - 2**(-0.5 / self.bands_per_octave))
        filters = 1 / (
            1
            + ((frequencies**2 - centers**2) / (frequencies * bandwidths))
            ** (2 * self.filter_order)
        )
        return filters



import numpy as np
import scipy.signal
from ._core import TimePeriod
import xarray as xr


class SampleTimer:
    def __init__(self, xr_obj):
        if 'samples' not in xr_obj.coords:
            raise TypeError('.sampling accessor only avaliable for xarrays with samples dimention')
        self._xr_obj = xr_obj

    @property
    def rate(self):
        return self._xr_obj.coords['samples'].attrs['rate']

    @property
    def num(self):
        return self._xr_obj.coords['samples'].size

    @property
    def time_period(self):
        start_time = self._xr_obj.coords['samples'].attrs['start_time']
        duration = self.num / self.rate
        return TimePeriod(start=start, duration=duration)

    def time_subperiod(self, time=None, /, *, start=None, stop=None, center=None, duration=None):
        original_period = self.time_period
        new_period = original_period.subperiod(time, start=start, stop=stop, center=center, duration=duration)
        start = (new_period.start - original_period.start).total_seconds()
        stop = (new_period.stop - original_period.start).total_seconds()
        # Indices assumed to be seconds from start
        start = np.math.floor(start * self.samplerate)
        stop = np.math.ceil(stop * self.samplerate)

        new_obj = self._xr_obj.isel(samples=slice(start, stop))
        new_obj.coords['samples'].attrs['start_time'] = self._start_time.add(seconds=start / self.samplerate)
        return new_obj


class Data:
    def __init__(self, data, dims=None, coords=None, attrs=None):
        if not isinstance(data, xr.DataArray):
            data = xr.DataArray(data, dims=dims, coords=coords, attrs=attrs)
        self._data = data

    def __add__(self, other):
        if isinstance(other, Data):
            other = other._data
        return self._remake(self._data + data)
        # new = type(self)(data=self._data + data)
        # self._transfer_attrs(new)
        # return new

    def reduce(self, function, dim, **kwargs):
        return self._remake(self._data.reduce(function, dim, **kwargs))
        # new = type(self)(data=self._data.reduce(function, dim, **kwargs))
        # self._transfer_attrs(new)
        # return new 

    def _remake(self, data):
        return Data(data=data)

    @property
    def data(self):
        import warnings
        warnings.warn('The data property should probably not be used', category=DeprecationWarning)
        return self._data.data

    # def _transfer_attrs(self, other):
    #     pass


class TimeData(Data):
    @classmethod
    def calibrated(cls, data, calibration, **kwargs):
        data = np.asarray(data)
        calibration = np.asarray(calibration).astype('float32')
        c = 10**(calibration / 20) / 1e-6  # Calibration values are given as dB re. 1μPa
        c = c.reshape((-1,) + (1,) * (data.ndim - 1))
        if data.dtype in (np.int8, np.int16, np.int32, np.float32):
            c = c.astype(np.float32)

        calibrated = data / c
        obj = cls(data=calibrated, **kwargs)
        return obj

    def __init__(self, data, start_time=None, samplerate=None, **kwargs):
        super().__init__(data=data, **kwargs)
        if 'start_time' not in self._data.attrs:
            if start_time is None:
                raise TypeError('Start time has to be given as input argument if it is missing from the xarray data')
            self._data = self._data.assign_attrs(start_time=start_time)
        if 'samplerate' not in self._data.attrs:
            if samplerate is None:
                raise TypeError('Samplerate has to be given as input argument if it is missing from the xarray data')
            self._data = self._data.assign_attrs(samplerate=samplerate)

    @property
    def samplerate(self):
        return self._data.samplerate

    @property
    def num_samples(self):
        return self._data.time.size

    @property
    def time_period(self):
        return TimePeriod(start=self._data.start_time, duration=self.num_samples / self.samplerate)

    def time_subperiod(self, time=None, /, *, start=None, stop=None, center=None, duration=None):
        original_period = self.time_period
        new_period = original_period.subperiod(time, start=start, stop=stop, center=center, duration=duration)
        start = (new_period.start - original_period.start).total_seconds()
        stop = (new_period.stop - original_period.start).total_seconds()
        # Indices assumed to be seconds from start
        start = np.math.floor(start * self.samplerate)
        stop = np.math.ceil(stop * self.samplerate)

        return self._remake(self._data.isel(time=slice(start, stop)).assign_attrs(start_time=self._start_time.add(seconds=start / self.samplerate)))

    def _remake(self, data):
        if 'time' in data.dims:
            return TimeData(data=data, samplerate=self.samplerate, start_time=self.start_time)
        return Data(data=data)

    # def _transfer_attrs(self, other):
    #     super()._transfer_attrs(other)
    #     if isinstance(other, TimeData):
    #         other._data = other._data.assign_attrs(samplerate=self.samplerate, start_time=self.start_time)
    
    # def reduce(self, function, dim, **kwargs):
    #     new_data = self._data.reduce(function, dim, **kwargs)
    #     if 'time' in new_data.dims:
    #         new = TimeData(data=new_data)  # Will break since you don't pass all arguments
    #     else:
    #         new = Data(data=new_data)
    #     self._transfer_attrs(new)
    #     return new


class FrequencyData(Data):
    def __init__(self, data, frequency=None, bandwidth=None, **kwargs):
        super().__init__(data, **kwargs)
        if 'frequency' not in self._data.coords:
            if frequency is None:
                raise TypeError('Frequency vector has to be given as input argument if it is missing from the xarray data')
            self._data = self._data.assign_coords(frequency=frequency)
        if 'bandwidth' not in self._data.attrs:
            if bandwidth is None:
                raise TypeError('Bandwidth has to be given as input argument if it is missing from the xarray data')
            self._data = self._data.assign_attrs(bandwidth=bandwidth)     

    def _remake(self, data):
        if 'frequency' in data.dims:
            return FrequencyData(data=data, frequency=self.frequency, bandwidth=self.bandwidth)
        return Data(data=data)


class TimeFrequencyData(TimeData, FrequencyData):
    def __init__(self, data, start_time=None, samplerate=None, frequency=None, bandwidth=None, **kwargs):
        super().__init__(data, start_time=start_time, samplerate=samplerate, frequency=frequency, bandwidth=bandwidth, **kwargs)

    def _remake(self, data):
        if 'time' in data.dims and 'frequency' in data.dims:
            return TimeFrequencyData(
                data=data,
                start_time=self.start_time,
                samplerate=self.samplerate,
                frequency=self.frequency,
                bandwidth=self.bandwidth
            )
        elif 'time' in data.dims:
            return TimeData(
                data=data,
                start_time=self.start_time,
                samplerate=self.samplerate,
            )
        elif 'frequency' in data.dims:
            return FrequencyData(
                data=data,
                frequency=self.frequency,
                bandwidth=self.bandwidth
            )
        else:
            return Data(data=data)


class Spectrogram(TimeFrequencyData):
    @classmethod
    def from_time_data(cls, time_data, window_duration, window='hann', overlap=0.5):
        if not isinstance(time_data, TimeData):
            raise TypeError(f"Cannot calculate the spectrogram of object of type '{time_data.__class__.__name__}'")
        window_samples = round(window_duration * time_data.samplerate)
        overlap_samples = round(window_duration * overlap * time_data.samplerate)
        f, t, Sxx = scipy.signal.spectrogram(
            x=time_data._data.data,
            fs=time_data.samplerate,
            window=window,
            nperseg=window_samples,
            noverlap=overlap_samples,
        )
        return cls(
            data=Sxx.copy(),  # Using a copy here is a performance improvement in later processing stages.
            # The array returned from the spectrogram function is the real part of the original stft, reshaped.
            # This means that the array takes twice the memory (the imaginary part is still around),
            # and it's not contiguous which slows down filtering a lot.
            samplerate=time_data.samplerate / (window_samples - overlap_samples),
            start_time=time_data.time_period.start.add(seconds=t[0]),
            frequency=f,
            bandwidth=time_data.samplerate / window_samples,
        )



import functools
import inspect


def prebind(func, **prebound_kwargs):
    sig = inspect.signature(func)

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        kwargs = prebound_kwargs | kwargs
        bound = sig.bind_partial(*args, **kwargs)
        bound.apply_defaults()
        if set(bound.arguments) == set(sig.parameters):
            return func(*args, **kwargs)
        else:
            wrapped = prebind(func, **bound.arguments)
            # Documentation of pre-bound values for human use
            if wrapped.__doc__:
                bound_docs = ['\n\nPre-bound parameters', '-' * 20]
                for key, value in bound.arguments.items():
                    bound_docs.append(f'{key}: {value}')
                wrapped.__doc__ += '\n'.join(bound_docs)
            # "Documentation" of pre-bound values for programmatic use
            wrapped._pre_bound = bound.arguments
            return wrapped

    return wrapper



#ruff: noqa
import uwacan
import numpy as np


class DummyRecording(uwacan.recordings.FileRecording):
    class RecordedFile(uwacan.recordings.FileRecording.RecordedFile):
        filepath = "dummy"
        def __init__(self, start_time, duration, samplerate, ref_time):
            self.filepath = f"{start_time}.dummy"
            self._start_time = start_time
            self._duration = duration
            self._samplerate = samplerate
            self.ref_time = ref_time

        @property
        def start_time(self):
            return self._start_time

        @property
        def stop_time(self):
            return self.start_time.add(seconds=self.duration)

        @property
        def duration(self):
            return self._duration

        @property
        def num_channels(self):
            return 1

        @property
        def num_samples(self):
            return int(self.duration * self.samplerate)

        @property
        def samplerate(self):
            return self._samplerate

        def read_data(self, start_idx=None, stop_idx=None):
            if start_idx is None:
                start_idx = 0
            if stop_idx is None:
                stop_idx = int(self.samplerate * self.duration)
            if start_idx < 0:
                raise ValueError("Cannot read file from before it started")
            if stop_idx > self.num_samples:
                raise ValueError("Cannot read file from after it ended")
            return np.arange(start_idx, stop_idx) / self.samplerate + (self.start_time - self.ref_time).total_seconds()

        def read_info(self):
            self._samplerate = self.__samplerate
            self._stop_time = self._start_time.add(seconds=self.__duration)

    @classmethod
    def make_dummy_recording(cls, num_files, file_duration, samplerate, ref_time="now", gaps=0):
        ref_time = uwacan._core.time_to_datetime(ref_time)
        files = [
            cls.RecordedFile(
                start_time=ref_time.add(seconds=idx * (file_duration + gaps)),
                duration=file_duration,
                samplerate=samplerate,
                ref_time=ref_time
            )
            for idx in range(num_files)
        ]
        return cls(files)

    def time_data(self):
        return uwacan.TimeData(self.raw_data(), start_time=self.time_window.start, samplerate=self.samplerate)
